# -*- coding: utf-8 -*-
"""
Helper script to configure user profile values
"""
from __future__ import print_function, division, absolute_import, unicode_literals
import os
import re
from io import open
from os import _exit as exit  # used for compatability with pyinstaller
from .exceptions import SDKConfigurationFileNotFound, SDKConfigurationProfileDoesNotExist

try:
    # for supporting python2
    input = raw_input
    FileExistsError = OSError
except NameError:
    pass
import configparser


CONFIG_FILEPATH = os.path.join(os.path.expanduser('~'), '.abeja', 'configx')


def load_abeja_config(profile='default'):
    if not os.path.exists(CONFIG_FILEPATH):
        raise SDKConfigurationFileNotFound('Configuration not set or missing, run "python -m abejaapi.configure" from the command line to configure!')
    config = configparser.ConfigParser()
    config.read(CONFIG_FILEPATH)
    try:
        profile_config = config[profile]
    except (KeyError, configparser.NoSectionError):
        raise SDKConfigurationProfileDoesNotExist('profile({}) Does not exist, use "python -m abejaapi.configure --profile PROFILE_NAME" to set new profile!'.format(profile))
    return profile_config


# TODO: add function without interactive ui
def is_valid_organization_name(name):
    if len(name) > 63:
        return False
    allowed = re.compile("(?!-)[A-Z\d-]{1,63}(?<!-)$", re.IGNORECASE)
    return allowed.match(name)


def set_configuration(profile='default'):
    config = configparser.ConfigParser(allow_no_value=True)
    config.read(CONFIG_FILEPATH)

    # get values from user
    config_keys = (('ABEJA_PLATFORM_SDK_USER_ID', None),
                   ('ABEJA_PLATFORM_SDK_USER_TOKEN', None),
                   ('ABEJA_PLATFORM_SDK_DATASOURCE_ID', None),
                   ('ABEJA_PLATFORM_SDK_DATASOURCE_TOKEN', None),
                   ('ABEJA_PLATFORM_SDK_ORGANIZATION_NAME', None),
                   ('ABEJA_PLATFORM_SDK_IAM_URL', 'api.iam.abeja.io'),
                   ('ABEJA_PLATFORM_SDK_API_URL', 'api.abeja.io'), )
    profile_config = {}
    confirmed = False
    print('ABEJA Platform SDK Configuration [{}]'.format(profile))
    print('-------------------------------------------')
    try:
        while not confirmed:
            for key, default in config_keys:
                if profile in config:
                    existing = config[profile].get(key, default)
                    default = existing
                while True:
                    result = input('{} [{}]: '.format(key, default)).strip()
                    if not result and default is not None:
                        result = default
                    if key == 'ABEJA_PLATFORM_SDK_ORGANIZATION_NAME' and not is_valid_organization_name(result):
                        print('ERROR: Invalid ORGANIZATION NAME, "{}"!'.format(result))
                        continue  # retry
                    elif key == 'ABEJA_PLATFORM_SDK_USER_ID' and not result.startswith('user-'):
                        print('ERROR: Invalid Value, user id, "{}", must start with "user-"!'.format(result))
                        continue  # retry
                    elif not result and not key.endswith(('DATASOURCE_ID', 'DATASOURCE_TOKEN')):  # allow datasource to be null
                        print('ERROR: Invalid Value, "{}"!'.format(result))
                        continue  # retry
                    profile_config[key] = result
                    break

            # confirmation
            # -- display
            print('------')
            print('Confirm:')
            for key, _ in config_keys:
                print('{}={}'.format(key, profile_config[key]))
            response = None
            while response not in ('y', 'n'):
                response = input('Is this correct? (Y/N) [Y]: ').strip().lower()
                if not response.strip():
                    response = 'y'
                if response == 'y':
                    confirmed = True
    except KeyboardInterrupt:
        print()  # ignore breaks
        exit(1)

    config[profile] = profile_config
    config_directory, _ = os.path.split(CONFIG_FILEPATH)
    try:
        os.mkdir(config_directory, 0o755)
    except FileExistsError:
        pass
    with open(CONFIG_FILEPATH, 'w', encoding='utf8') as output_fileobj:
        config.write(output_fileobj)
    print('Configuration written:')
    print(CONFIG_FILEPATH)
    exit(0)


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('-p', '--profile',
                        default='default',
                        help='If given, the values given will be assigned to this profile name')
    args = parser.parse_args()
    set_configuration(args.profile)
