# -*- coding: utf-8 -*-
"""
Core components of the abejaapi sdk
"""
from __future__ import print_function, division, absolute_import, unicode_literals
import os
import logging
import random
from time import sleep
try:
    from functools import lru_cache
except ImportError:
    from .backports import lru_cache

import requests
from . import __version__ as VERSION
from .configure import load_abeja_config
from .exceptions import (
    InvalidCredentials,
    OrganizationIdsNotDefined,
    UnknownChannelId,
    Http403Forbidden,
    UnknownError,
    InvalidOrganization,
    InvalidChannelStorageType,
    SDKConfigurationFileNotFound,
    SDKConfigurationProfileDoesNotExist,
    SDKConfigurationFileCorrupt,
)

PACKAGE_LOGGER_NAME = 'abejaapi'

# configure logging
logging.basicConfig(level=logging.INFO,
                    format='%(relativeCreated)6d %(threadName)s %(message)s')
logger = logging.getLogger(PACKAGE_LOGGER_NAME)

# Process ENVIRONMENT Variables
ABEJA_PLATFORM_SDK_USER_ID = os.environ.get('ABEJA_PLATFORM_SDK_USER_ID', None)
ABEJA_PLATFORM_SDK_USER_TOKEN = os.environ.get('ABEJA_PLATFORM_SDK_USER_TOKEN', None)
ABEJA_PLATFORM_SDK_DATASOURCE_ID = os.environ.get('ABEJA_PLATFORM_SDK_DATASOURCE_ID', None)
ABEJA_PLATFORM_SDK_DATASOURCE_TOKEN = os.environ.get('ABEJA_PLATFORM_SDK_DATASOURCE_TOKEN', None)
ABEJA_PLATFORM_SDK_ORGANIZATION_NAME = os.environ.get('ABEJA_PLATFORM_SDK_ORGANIZATION_NAME', None)
IAM_BASE_URL = os.environ.get('ABEJA_PLATFORM_SDK_IAM_URL', 'api.iam.abeja.io')
API_BASE_URL = os.environ.get('ABEJA_PLATFORM_SDK_API_URL', 'api.abeja.io')

DEFAULT_CHANNEL_STORAGETYPE = 'datalake'
VALID_CHANNEL_STORAGETYPES = (
    'datalake',
    'file',
    'rdb',
)

if not any((ABEJA_PLATFORM_SDK_USER_ID, ABEJA_PLATFORM_SDK_USER_TOKEN)):
    profile = os.environ.get('ABEJA_PLATFORM_SDK_PROFILE', 'default')
    if not profile:
        profile = 'default'
    logger.debug('PROFILE:{}'.format(profile))
    try:
        config = load_abeja_config(profile)
        ABEJA_PLATFORM_SDK_USER_ID = config.get('ABEJA_PLATFORM_SDK_USER_ID', None)
        ABEJA_PLATFORM_SDK_USER_TOKEN = config.get('ABEJA_PLATFORM_SDK_USER_TOKEN', None)
        ABEJA_PLATFORM_SDK_DATASOURCE_ID = config.get('ABEJA_PLATFORM_SDK_DATASOURCE_ID', None)
        ABEJA_PLATFORM_SDK_DATASOURCE_TOKEN = config.get('ABEJA_PLATFORM_SDK_DATASOURCE_TOKEN', None)
        ABEJA_PLATFORM_SDK_ORGANIZATION_NAME = config.get('ABEJA_PLATFORM_SDK_ORGANIZATION_NAME', None)
        IAM_BASE_URL = config.get('ABEJA_PLATFORM_SDK_IAM_URL', 'api.iam.abeja.io')
        API_BASE_URL = config.get('ABEJA_PLATFORM_SDK_API_URL', 'api.abeja.io')
    except (SDKConfigurationFileNotFound, SDKConfigurationProfileDoesNotExist) as e:
        logger.debug('No profile found, configuration not used to set values: {}'.format(e.args))
    except KeyError as e:
        raise SDKConfigurationFileCorrupt('Missing Expected Key, delete ~/.abeja/configx and re-configure: {}'.format(e.args))


SUCCESSFUL_STATUS_CODES = range(200, 300)  # 2xx Success HTTP Status Codes [200-299]
USER_ID_PREFIX = 'user-'
ABEJA_METADATA_PREFIX = 'x-abeja-meta-'


def retry(tries=3):
    """
    Exponential Backoff retry decorator
    """
    if tries <= 0:
        raise ValueError('tries must be > 0!')

    def decorator(f):
        def retry_function(*args, **kwargs):
            result = None
            for attempt in range(tries):
                try:
                    result = f(*args, **kwargs)
                    break  # success
                except Exception as e:
                    if attempt == tries - 1:  # last attempt
                        raise  # re-raise exception
                    random_milliseconds = random.uniform(0, 1)
                    seconds = ((2 ** attempt) + random_milliseconds)
                    logger.warning('{} occurred, retrying after {}s'.format(e.__class__.__name__, seconds))
                    sleep(seconds)
            return result
        return retry_function
    return decorator


class PlatformResourceManager(object):

    def __init__(self, user_id=ABEJA_PLATFORM_SDK_USER_ID, user_token=ABEJA_PLATFORM_SDK_USER_TOKEN, organization=ABEJA_PLATFORM_SDK_ORGANIZATION_NAME, iam_base_url=IAM_BASE_URL):
        if not (user_id and user_token):
            raise InvalidCredentials('(user_id|token) NOT given or set!')
        if not user_id.startswith(USER_ID_PREFIX):
            raise InvalidCredentials('Given UserId({}), does not startwith: "{}"'.format(user_id,
                                                                                         USER_ID_PREFIX))
        self._user_id = user_id
        self._user_token = user_token
        self._user_credentials = (self._user_id, self._user_token)
        self._iam_base_url = iam_base_url
        self._session = requests.session()
        self._session.headers.update({'accept': 'application/json'})
        self._session.headers.update({'User-Agent': 'abejaapi-sdk-{}'.format(VERSION)})

        self._organization_name = organization
        self._organization_id = None
        self._my_user_id_cache = None
        self._my_organization_ids_cache = [] 

    @property
    def _my_user_id(self):
        if not self._my_user_id_cache:
             self._populate_my_organizations()  # populates self._my_user_id_cache
        return self._my_user_id_cache

    @property
    def _my_organization_ids(self):
        if not self._my_organization_ids_cache:
            self._populate_my_organizations()  # populates self._my_organization_ids_cache
        return self._my_organization_ids_cache
 
    def _populate_my_organizations(self):
        my_info = self.me()
        logger.debug('Setting _my_user_id...')
        self._my_user_id_cache = my_info['id']
        organization_data = self.organizations()
        for organization_info in organization_data['organizations']:
            organization_id = organization_info['id']
            self._my_organization_ids_cache.append(organization_id)

    def me(self):
        """        
        :return: (dict) result of /users/me 
        """
        url = 'https://{}/users/me'.format(self._iam_base_url)
        logger.debug('Request URL: {}'.format(url))
        response = self._session.get(url, auth=self._user_credentials)
        logger.debug('/users/me Response: ({}) {}'.format(response.status_code,
                                                          response.text))
        if response.status_code == 403:
            content = response.json()
            message = content.get('Message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('StatusCode({}) not in {}: {}'.format(response.status_code,
                                                                     SUCCESSFUL_STATUS_CODES,
                                                                     response.text))
        return response.json()

    def set_organization(self, organization_name):
        # get the id for the given org
        organization_id = None
        existing_names = []
        for org_info in self.organizations()['organizations']:
            if organization_name == org_info['name']:
                organization_id = org_info['id']
                break
            existing_names.append(org_info['name'])
        if not organization_id:
            raise InvalidOrganization('({}) not in {}'.format(organization_name, existing_names))
        self._organization_id = organization_id
        self._organization_name = organization_name

    def add_channel(self, organization_id=None, name=None, type=DEFAULT_CHANNEL_STORAGETYPE):
        """
        A channel is a core entry-point unit in the ABEJA platform.
        A channel is necessary for linking to a Datasource
        :return: ADD_CHANNEL_RESPONSE
        """
        if not name:
            # generate random 10 digit name
            name = 'channel-name-{}'.format(''.join(str(random.randint(0, 9)) for _ in range(10)))

        if type not in VALID_CHANNEL_STORAGETYPES:
            raise InvalidChannelStorageType('Given StorageType({}) not in: {}'.format(type, VALID_CHANNEL_STORAGETYPES))

        if not organization_id and len(self._my_organization_ids) == 1:
            organization_id = self._my_organization_ids[0]
        elif self._organization_id:
            organization_id = self._organization_id

        if not organization_id:
            raise OrganizationIdsNotDefined('User Account({}) has no associated organizations, or organization_ids > 1: {}'.format(self._my_user_id,
                                                                                                                                   self._my_organization_ids))

        logger.debug('Using Organization({})...'.format(organization_id))

        url = 'https://{}/organizations/{}/channels'.format(self._iam_base_url,
                                                            organization_id)
        payload = {
            'display_name': name,
            'storage_type': type,
        }
        logger.debug('Request URL: {}'.format(url))
        logger.debug('Adding Channel({})...'.format(name))
        response = self._session.post(url,
                                      json=payload,
                                      auth=self._user_credentials)
        logger.debug('response ({}): {}'.format(response.status_code,
                                                response.text))
        if response.status_code == 403:
            content = response.json()
            message = content.get('Message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('StatusCode({}) not in {}: {}'.format(response.status_code,
                                                                     SUCCESSFUL_STATUS_CODES,
                                                                     response.text))
        self.channels.cache_clear()  # clear lru_cache on channel add
        self.channel_datasources.cache_clear()
        return response.json()

    @lru_cache(maxsize=5)
    def channels(self, organization_id=None):
        """        
        :param organization_id: (int) If given, only results for this organization will be provided.
        :return: (dict) {'channels': [{"id":"XXX","organization_id":"XXX","display_name":"XXX","created_at":"2017-03-01T12:44:09Z","updated_at":"2017-03-04T06:09:32Z"}, ...], }
        """
        if organization_id:
            organization_ids = [organization_id]
        else:
            organization_ids = self._my_organization_ids

        if not organization_ids:
            raise OrganizationIdsNotDefined('User Account({}) has no associated organizations, and no organization_id given!'.format(self._my_user_id))

        all_channels = {'channels': []}
        for organization_id in organization_ids:
            logger.debug('Processing Organization({})...'.format(organization_id))
            url = 'https://{}/organizations/{}/channels'.format(self._iam_base_url,
                                                                organization_id)
            response = self._session.get(url,
                                         auth=self._user_credentials)
            if response.status_code == 403:
                content = response.json()
                message = content.get('Message', 'Reason Unknown')
                raise Http403Forbidden(message)
            elif response.status_code not in SUCCESSFUL_STATUS_CODES:
                raise UnknownError('StatusCode({}) not in {}: {}'.format(response.status_code,
                                                                         SUCCESSFUL_STATUS_CODES,
                                                                         response.text))
            # flattening results by combining multiple organizations in {'channels': [...]}
            organization_channels = response.json()['channels']
            # WARNING -- This section adds a attribute to channel_info that is NOT originally available!!!! (... to make life easier else where)
            for channel in organization_channels:
                channel['organization_id'] = organization_id  # Adding Organization ID to channel info
            all_channels['channels'].extend(organization_channels)
        return all_channels

    def add_datasource(self, organization_id=None, name=None, channel_id=None):
        """
        :param channel_id: (int) CHANNEL_ID of channel to link datasource to
        :return: (tuple)  (ADD_DATASOURCE_RESPONSE, LINK_CHANNEL_RESPONSE)
        """
        if not name:
            # generate random 10 digit name
            name = 'datasource-name-{}'.format(''.join(str(random.randint(0, 9)) for _ in range(10)))

        if not organization_id and len(self._my_organization_ids) == 1:
            organization_id = self._my_organization_ids[0]
        if not organization_id:
            raise OrganizationIdsNotDefined('User Account({}) has no associated organizations, or organization_ids > 1: {}'.format(self._my_user_id,
                                                                                                                                   self._my_organization_ids))

        url = 'https://{}/organizations/{}/datasources'.format(self._iam_base_url,
                                                               organization_id)
        payload = {
            'organization_id': organization_id,
            'display_name': name,
        }
        logger.debug('Request URL: {}'.format(url))
        logger.info('Creating new Datasource({}) in Organization({})...'.format(name,
                                                                                organization_id))
        add_datasource_response = self._session.post(url,
                                                     json=payload,
                                                     auth=self._user_credentials)
        if add_datasource_response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('StatusCode({}) not in {}: {}'.format(add_datasource_response.status_code,
                                                                     SUCCESSFUL_STATUS_CODES,
                                                                     add_datasource_response.text))
        datasource = add_datasource_response.json()
        channel_link = None
        if channel_id:
            datasource_id = datasource['datasource']['datasource_id']
            logger.info('Linking DataSource({}) to Channel({})...'.format(datasource_id,
                                                                          channel_id))
            url = 'https://{}/organizations/{}/channels/{}/datasources/{}'.format(self._iam_base_url,
                                                                                  organization_id,
                                                                                  channel_id,
                                                                                  datasource_id)
            channel_link_response = self._session.put(url,
                                                      auth=self._user_credentials)
            logger.debug('Response ({}): {}'.format(channel_link_response.status_code,
                                                    channel_link_response.text))
            if channel_link_response.status_code not in SUCCESSFUL_STATUS_CODES:
                raise UnknownError('StatusCode({}) not in {}: {}'.format(channel_link_response.status_code,
                                                                         SUCCESSFUL_STATUS_CODES,
                                                                         channel_link_response.text))
            channel_link = channel_link_response.json()
        self.datasources.cache_clear()
        self.channel_datasources.cache_clear()
        return datasource, channel_link

    @lru_cache(maxsize=5)
    def datasources(self, organization_id=None):
        """        
        :param organization_id: (iterator) If given, only information for given Datasource IDs will be returned (ex: [29820982098892, 1198891981, ...] 
        :return: (tuple)  
            ((DATASOURCE_ID, DATASOURCE_TOKEN, DATASOURCE_NAME, ORGANIZATION_ID, (CHANNEL_ID, CHANNEL_ID, ...), ...) 
        """
        if organization_id:
            organization_ids = [organization_id]
        else:
            organization_ids = self._my_organization_ids

        if not organization_ids:
            raise OrganizationIdsNotDefined('User Account({}) has no associated organizations, and no organization_id given!'.format(self._my_user_id))

        all_datasources = []
        for organization_id in organization_ids:
            url = 'https://{}/organizations/{}/datasources'.format(self._iam_base_url,
                                                                   organization_id)
            response = self._session.get(url,
                                         auth=self._user_credentials)
            if response.status_code == 403:
                content = response.json()
                message = content.get('Message', 'Reason Unknown')
                raise Http403Forbidden(message)
            elif response.status_code not in SUCCESSFUL_STATUS_CODES:
                raise UnknownError('StatusCode({}) not in {}: {}'.format(response.status_code,
                                                                         SUCCESSFUL_STATUS_CODES,
                                                                         response.text))
            organization_channels = response.json()['datasources']  # flattening results...
            all_datasources.extend(organization_channels)
        return all_datasources

    @lru_cache(maxsize=50)
    def channel_datasources(self, channel_id, organization_id=None):
        """
        Collect the datasources that are assigned to a given channel.
        :param channel_id: (int) CHANNEL_ID
        :param organization_id: (int) [OPTIONAL] If not provided the user's Organization will be used if SINGULAR
        :return: (list) Datasources belonging to the given Channel
        """
        if not organization_id and len(self._my_organization_ids) == 1:
            organization_id = self._my_organization_ids[0]

        logger.debug(self._my_organization_ids)
        if not organization_id:
            # search organizations for given channel_id to get the related organization_id
            channels = self.channels()
            for channel_info in channels['channels']:
                candidate_organization_id = channel_info['organization_id']
                candidate_channel_id = channel_info['channel_id']
                if candidate_channel_id == channel_id:
                    organization_id = candidate_organization_id
                    break

        if not organization_id:
            raise UnknownChannelId('Given ChannelId({}) not found in User Organizations: {}'.format(channel_id,
                                                                                                    self._my_organization_ids))
        url = 'https://{}/organizations/{}/channels/{}/datasources'.format(self._iam_base_url,
                                                                           organization_id,
                                                                           channel_id)
        response = self._session.get(url,
                                     auth=self._user_credentials)
        if response.status_code == 403:
            content = response.json()
            message = content.get('Message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('StatusCode({}) not in {}: {}'.format(response.status_code,
                                                                     SUCCESSFUL_STATUS_CODES,
                                                                     response.text))
        return response.json()

    @lru_cache(maxsize=5)
    def organizations(self):
        """
        :return: (list) list of organization objects
        """
        url = 'https://{}/users/{}/organizations'.format(self._iam_base_url,
                                                         self._my_user_id)
        logger.debug('Request URL: {}'.format(url))
        response = self._session.get(url,
                                     auth=self._user_credentials)
        if response.status_code == 403:
            content = response.json()
            message = content.get('Message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('StatusCode({}) not in {}: {}'.format(response.status_code,
                                                                     SUCCESSFUL_STATUS_CODES,
                                                                     response.text))
        return response.json()

