"""
The classes defined here in agents.py are intended for use with in a 'datasource' component.
"""
import os
import datetime
import logging
import hashlib

import requests

from ..core import ABEJA_PLATFORM_SDK_DATASOURCE_ID, ABEJA_PLATFORM_SDK_DATASOURCE_TOKEN, SUCCESSFUL_STATUS_CODES, ABEJA_METADATA_PREFIX, API_BASE_URL
from ..exceptions import InvalidCredentials, Http403Forbidden, Http401Unauthorized, UnknownError
from .exceptions import DatasourceUserNotSet, InvalidDatasourceUser

logger = logging.getLogger('abejaapi')

DATASOURCE_ID_PREFIX = 'datasource-'

UNKNOWN_CONTENTTYPE_DEFAULT = 'binary/octet-stream'
FILE_EXTENSION_CONTENTTYPE_MAPPING = {
    '.csv': 'text/csv',
    '.json': 'application/json',
    '.xml': 'application/xml',
    '.gif': 'image/gif',
    '.jpg': 'image/jpeg',
    '.jpeg': 'image/jpeg',
    '.pdf': 'application.pdf',
    '.mpeg': 'video/mpeg',
}


class PlatformDatasourceAgent(object):
    """
    This Agent is for implementing a simple Datasource for sending data to the ABEJA Platform
    """

    def __init__(self, datasource_id=ABEJA_PLATFORM_SDK_DATASOURCE_ID, datasource_token=ABEJA_PLATFORM_SDK_DATASOURCE_TOKEN, api_base_url=API_BASE_URL):
        self._datasource_id = datasource_id
        self._datasource_token = datasource_token
        if not (datasource_id and datasource_token):
            raise InvalidCredentials('(username|token) NOT given or set!')

        self._datasource_credentials = (self._datasource_id, self._datasource_token)
        self._api_base_url = api_base_url
        self._session = requests.session()
        self._session.headers.update({'accept': 'application/json'})

        self._my_user_id = None
        self._uploaded_files = []

    def _determine_contenttype(self, filename):
        """
        Determine the HTTP 'Content-Type' header value to use from the given filename extension.
        If file extension is unknown UNKNOWN_CONTENTTYPE_DEFAULT will be used.
        :param filename:
        :return:
        """
        _, extension = os.path.splitext(filename)
        contenttype = FILE_EXTENSION_CONTENTTYPE_MAPPING.get(extension, UNKNOWN_CONTENTTYPE_DEFAULT)
        logger.debug('Using Content-Type: {}'.format(contenttype))
        return contenttype

    def _get_datasource_id(self, datasource_id):
        """
        Extract the DATASOURCE_ID from datasource user name
        :param datasource_id: (str) DATASOURCE_ID (ex: datasource-NNNNNNNNNNNNN)
        :return: (int) NNNNNNNNNNNNN
        """
        if not self._datasource_id.startswith(DATASOURCE_ID_PREFIX):
            raise InvalidCredentials('Given Datasource User({}), does not startwith: "{}"'.format(datasource_id,
                                                                                                  DATASOURCE_ID_PREFIX))
        prefix, datasource_id = datasource_id.split('-')
        return int(datasource_id)

    def _get_signed_url(self, datasource_id, datasource_token, channel_id, contenttype, metadata):
        """
        Obtain the SIGNED URL needed for file upload.
        :param datasource_id: (str) DATASOURCE ID
        :param datasource_token: (str) DATASOURCE TOKEN
        :param channel_id: CHANNEL ID to request URL for
        :param contenttype:
        :param metadata: (dict) Any user defined metadata to apply on upload
        :return: SIGNED_URL, FILE_ID
        """
        assert channel_id
        credentials = (datasource_id, datasource_token)
        url = 'https://{}/channels/{}'.format(self._api_base_url, channel_id)

        # Update headers
        self._session.headers.update({'content-type': contenttype})

        # add metadata headers
        self._session.headers.update(metadata)
        logger.debug('_get_signed_url HEADERS: {}'.format(self._session.headers))

        logger.debug('Requesting signed URL: {}'.format(url))
        response = self._session.post(url,
                                      auth=credentials)
        logger.debug('Response ({}): {}'.format(response.status_code, response.text))
        if response.status_code == 403:
            content = response.json()
            message = content.get('message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code == 401:
            content = response.json()
            message = content.get('message', 'Reason Unknown')
            raise Http401Unauthorized(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise Exception('{} {}'.format(response.status_code, response.text))

        request_url_data = response.json()
        file_id = request_url_data['file_id']
        signed_url = request_url_data['upload_url']
        return signed_url, file_id

    def upload_file(self, local_filepath, channel_id, metadata=None, user=None, token=None):
        """
        Uploads a local file to the given CHANNEL
        (a light wrapper around upload_fileobj)
        :param local_filepath:
        :param channel_id: (int) CHANNEL Identifier
        :param metadata: (dict) Metadata to apply to the file (fieldname may/may not include the ABEJA_METADATA_PREFIX)
        :param user: (str) [OPTIONAL] DATASOURCE_USER (if not given, value given on instantiation will be used)
        :param token: (str) [OPTIONAL] DATASOURCE_TOKEN (if not given, value given on instantiation will be used)
        :return: (str) FILE_ID (used for retriveing addtional file information and downloading the file)
        """
        logger.debug('Opening file({})...'.format(local_filepath))
        with open(local_filepath, 'rb') as fileobj:
            file_id = self.upload_fileobj(fileobj, channel_id, metadata, user, token)
        return file_id

    def _get_fileobj_md5(self, obj):
        """
        Determine the MD5 hash of a given file object
        :param obj: (file-like object)
        :return: (str) MD5 hexdigest
        """
        obj_hash = hashlib.md5(obj.read())
        obj.seek(0)  # reset obj file read position
        return obj_hash.hexdigest()

    def upload_fileobj(self, obj, channel_id, metadata=None, datasource_id=None, datasource_token=None):
        """
        Uploads a python file-like object to the given CHANNEL
        NOTE: The following metadata is automatically added on upload by this method:
            - x-abeja-meta-filename
            - x-abeja-meta-md5

        :param obj: (file object) File Object to upload
        :param channel_id: (int) CHANNEL Identifier
        :param metadata: (dict) Metadata to apply to the file (fieldname may/may not include the ABEJA_METADATA_PREFIX)
        :param datasource_id: (str) [OPTIONAL] DATASOURCE_ID (if not given, value given on instantiation will be used)
        :param token: (str) [OPTIONAL] DATASOURCE_TOKEN (if not given, value given on instantiation will be used)
        :return: (str) FILE_ID (used for retriveing additional file information and downloading the file)
        """
        if not datasource_id:
            datasource_id = self._datasource_id
        if not datasource_token:
            datasource_token = self._datasource_token

        if not all((id, datasource_token)):
            msg = 'ABEJA_PLATFORM_SDK_DATASOURCE_ID"/"ABEJA_PLATFORM_SDK_DATASOURCE_TOKEN USER/TOKEN not set or given!"'
            raise DatasourceUserNotSet(msg)

        if not datasource_id.startswith(DATASOURCE_ID_PREFIX):
            raise InvalidDatasourceUser('id({}) does not start with expected, "{}" prefix!'.format(datasource_id,
                                                                                                   DATASOURCE_ID_PREFIX))
        normalized_metadata = {}
        if metadata:
            for fieldname, v in metadata.items():
                normalized_fieldname = self._normalize_metadata_fieldname(fieldname)
                normalized_metadata[normalized_fieldname] = str(v)  # HEADER value MUST BE STRING/BYTES not int/float

        # Add md5hash as metadata
        md5hash = self._get_fileobj_md5(obj)
        metadata_md5_fieldname = '{}md5'.format(ABEJA_METADATA_PREFIX)
        normalized_metadata[metadata_md5_fieldname] = md5hash

        # Add filename as metadata
        filename = os.path.split(obj.name)[-1]  # get filename without local path!
        metadata_filename_fieldname = '{}filename'.format(ABEJA_METADATA_PREFIX)
        normalized_metadata[metadata_filename_fieldname] = filename

        contenttype = self._determine_contenttype(obj.name)
        signed_url, file_id = self._get_signed_url(datasource_id, datasource_token, channel_id, contenttype, normalized_metadata)

        # add headers
        self._session.headers.update({'content-type': contenttype})
        metadata_filename_fieldname = '{}filename'.format(ABEJA_METADATA_PREFIX)
        self._session.headers.update({metadata_filename_fieldname: filename})  # Add filename metadata
        logger.debug('upload_fileobj HEADERS: {}'.format(self._session.headers))

        # add metadata headers
        self._session.headers.update(normalized_metadata)

        # send binary file via PUT to signed URL
        logger.debug('PUT {} to: {}'.format(obj.name, signed_url))
        response = self._session.put(signed_url,
                                     data=obj.read())
        logger.debug('Response ({}): {}'.format(response.status_code, response.text))
        if response.status_code == 403:
            content = response.json()
            message = content.get('Message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('StatusCode({}) not in {}: {}'.format(response.status_code,
                                                                     SUCCESSFUL_STATUS_CODES,
                                                                     response.text))
        # cache uploaded uploaded file info (not needed??)
        signed_url_issue_datetime = datetime.datetime.now()
        self._uploaded_files.append((filename,
                                     file_id,
                                     signed_url_issue_datetime,
                                     signed_url))
        return file_id

    def _normalize_metadata_fieldname(self, fieldname):
        """
        Given a metadata fieldname, add prefix, 'x-abeja-meta-', if missing

        :param fieldname:
        :return:
        """
        normalized_fieldname = fieldname.lower()
        if not normalized_fieldname.startswith(ABEJA_METADATA_PREFIX):
            # add 'x-abeja-meta-'
            normalized_fieldname = '{}{}'.format(ABEJA_METADATA_PREFIX, normalized_fieldname)
        return normalized_fieldname
