# -*- coding: utf-8 -*-
"""
abeja platform upload functions
"""
from __future__ import print_function, division, absolute_import, unicode_literals
import datetime
import os
import logging
import tempfile
from multiprocessing import cpu_count
from multiprocessing.pool import ThreadPool as Pool
import pytz
import requests
import dateutil.parser

from .exceptions import Md5HashDoesNotMatch
from .agents import PlatformDatasourceAgent, ABEJA_PLATFORM_SDK_DATASOURCE_ID, ABEJA_PLATFORM_SDK_DATASOURCE_TOKEN, DATASOURCE_ID_PREFIX
from ..exceptions import Http403Forbidden, Http401Unauthorized, UnknownError
from ..core import PlatformResourceManager, retry, API_BASE_URL, SUCCESSFUL_STATUS_CODES

logger = logging.getLogger('abejaapi')

ABEJA_METADATA_PREFIX = 'x-abeja-meta-'
ABEJA_UPLOADED_AT_EPOCH_UTC = datetime.datetime(2017, 5, 1, tzinfo=pytz.utc)
REQUEST_PROCESSES = cpu_count() * 2

ONE_MEGABYTE = 1024 * 1024
DOWNLOAD_CHUNK_SIZE_BYTES = ONE_MEGABYTE 


@retry(tries=3)
def process_datalake_list_request(args):
    """
    Process a DATALAKE API list request
    :param args: (session, user_credentials, url, start_date, end_date)
        NOTE: start_date & end_date are in UTC
    :return: (args, response.url, pages)
    """
    session, credentials, url, start_date, end_date = args
    pages = []

    is_initial = True
    next_page_token = None
    file_info = None

    page = []  # just for the initial while statement... probably not needed..
    while is_initial or ('next_page_token' in page and next_page_token):
        params = {
            'timezone': 'UTC',
            'start': start_date.strftime('%Y%m%d'),  # YYYYMMDD
            'end': end_date.strftime('%Y%m%d')  # YYYYMMDD
        }

        response = session.get(url,
                               params=params,
                               auth=credentials)
        if response.status_code == 403:
            content = response.json()
            message = content.get('Message', 'Reason Unknown')
            raise Http403Forbidden(message)
        logger.debug(response.status_code)
        if response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('StatusCode({}) not in {}: {}'.format(response.status_code,
                                                                     SUCCESSFUL_STATUS_CODES,
                                                                     response.text))
        page = response.json()
        # update "files" 'uploaded_at' attribute
        for file_info in page['files']:
            file_info['uploaded_at'] = dateutil.parser.parse(file_info['uploaded_at'])
            file_info['url_expires_on'] = dateutil.parser.parse(file_info['url_expires_on'])

            # temporary fix until DATALAKE supports proper int/float handling
            # Convert metadata string values to int/float in order to properly support gte/lte conditionals
            for k, v in file_info['metadata'].items():
                try:
                    v = int(v)
                except ValueError:
                    try:
                        v = float(v)
                    except ValueError:
                        pass
                file_info['metadata'][k] = v
        pages.append(page)
        is_initial = False
        if file_info:
            next_page_token = file_info.get('next_page_token', None)
        else:
            next_page_token = None
    return args, url, pages


class PlatformDataManager(PlatformResourceManager, PlatformDatasourceAgent):
    """
    Provides features for interacting with data on the ABEJA Platform
    """

    def __init__(self, *args, **kwargs):
        # capture potentially set variables
        self._datasource_id = kwargs.pop('datasource_id', ABEJA_PLATFORM_SDK_DATASOURCE_ID)
        self._datasource_token = kwargs.pop('datasource_token', ABEJA_PLATFORM_SDK_DATASOURCE_TOKEN)
        self._datasource_credentials = (self._datasource_id, self._datasource_token)
        self._api_base_url = kwargs.pop('api_base_url', API_BASE_URL)
        self._uploaded_files = []

        if self._datasource_id:
            assert self._datasource_id.startswith(DATASOURCE_ID_PREFIX)

        super(PlatformDataManager, self).__init__(*args, **kwargs)

    @retry(tries=3)
    def update_metadata(self, channel_id, file_id, metadata):
        """
        Update the metadata attached to a given file_id.

        :param channel_id: (int) CHANNEL_ID
        :param file_id: (str) File Identifier issued to the file (See list_channel)
        :param metadata: (dict) Key-Value mapping of metadata to attach to the file (existing metadata keys will be overwritten)
            NOTE: The ABEJA metadata prefix, 'x-abeja-meta-' will be added if missing.
        :return: (dict) Response body
        """
        # normalize metadata
        normalized_metadata = {}
        for fieldname, value in metadata.items():
            normalized_fieldname = self._normalize_metadata_fieldname(fieldname)
            normalized_metadata[normalized_fieldname] = value
        logger.debug('update_metadata() normalized_metadata: {}'.format(normalized_metadata))

        # get url using file_id
        file_info_url = 'https://{}/channels/{}/{}/metadata'.format(self._api_base_url, channel_id, file_id)
        response = requests.put(file_info_url,
                                auth=self._datasource_credentials,
                                json=normalized_metadata)

        logger.debug('Response ({}): {}'.format(response.status_code, response.text))
        if response.status_code == 403:
            content = response.json()
            message = content.get('message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code == 401:
            content = response.json()
            message = content.get('message', 'Reason Unknown')
            raise Http401Unauthorized(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('StatusCode({}) not in {}: {}'.format(response.status_code,
                                                                     SUCCESSFUL_STATUS_CODES,
                                                                     response.text))
        file_info = response.json()
        return file_info

    def download_bytes(self, file_id, channel_id, chunk_size=DOWNLOAD_CHUNK_SIZE_BYTES):
        """
        Download file identified by the file_id to bytes
        NOTE: Download uses the USER credentials (Not Datasource credentials)

        :param file_id: FILE_ID of file to download
        :param channel_id: CHANNEL_ID from which to download file from
        :param chunk_size: Chunk Size in bytes
        :return: ((str), (iter)) filename, bytes of target file in chunk_size
        """
        # get url using file_id
        file_info_url = 'https://{}/channels/{}/{}'.format(self._api_base_url, channel_id, file_id)
        response = self._session.get(file_info_url,
                                     auth=self._user_credentials)
        logger.debug('Response ({}): {}'.format(response.status_code, response.text))
        if response.status_code == 403:
            content = response.json()
            message = content.get('message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('StatusCode({}) not in {}: {}'.format(response.status_code,
                                                                     SUCCESSFUL_STATUS_CODES,
                                                                     response.text))
        file_info = response.json()
        download_url = file_info['download_url']

        # *REMOVE* session headers (Temporary Signed URL usage cannot be used when headers are given!!!)
        temp_headers_store = self._session.headers.copy()
        self._session.headers = {}
        response = self._session.get(download_url,
                                     stream=True)
        logger.debug('Response ({}): {}'.format(response.status_code, response.text))
        if response.status_code == 403:
            content = response.json()
            message = content.get('message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('StatusCode({}) not in {}: {}'.format(response.status_code,
                                                                     SUCCESSFUL_STATUS_CODES,
                                                                     response.text))
        # return headers
        self._session.headers.update(temp_headers_store)

        return file_info, response.iter_content(chunk_size=chunk_size)

    @retry(tries=3)
    def download_file(self, file_id, channel_id, local_filepath=None, chunk_size=DOWNLOAD_CHUNK_SIZE_BYTES):
        """
        Download file identified by the file_id
        NOTE: Download uses the USER credentials (Not Datasource credentials)

        :param file_id: FILE_ID of file to download or TEMPORARY_SIGNED_URL
        :param channel_id: CHANNEL_ID from which to download file from
        :param local_filepath: [OPTIONAL] If given, file will be placed here
        :param chunk_size: Chunk Size in bytes
        :return: (str) local file path to downloaded file 
        """
        file_info, download_bytes_iter = self.download_bytes(file_id, channel_id, chunk_size)
        metadata_filename_fieldname = '{}filename'.format(ABEJA_METADATA_PREFIX)
        filename = file_info['metadata'].get(metadata_filename_fieldname, None)

        metadata_md5_fieldname = '{}md5'.format(ABEJA_METADATA_PREFIX)
        file_md5hash = file_info['metadata'].get(metadata_md5_fieldname, None)

        # download file!
        if not local_filepath:
            local_filepath = tempfile.gettempdir()

        if not filename:
            filename = os.path.basename(
                tempfile.NamedTemporaryFile(prefix='abeja-datalake-').name)  # get random filename
        local_filepath = os.path.join(local_filepath, filename)

        with open(local_filepath, 'wb') as f:
            # download as stream & write in chunks to reduce memory load
            for chunk in download_bytes_iter:
                f.write(chunk)

        if file_md5hash:
            logger.debug('Performing Md5 check...')
            with open(local_filepath, 'rb') as f:
                actual_md5hash = self._get_fileobj_md5(f)
                if actual_md5hash != file_md5hash:
                    raise Md5HashDoesNotMatch('{} hash does not match!'.format(local_filepath))
        return local_filepath

    def list_channel(self, channel_id, start_date=None, end_date=None, timezone=None):
        """
        Get a list of items in the given channel

        :param channel_id: (int) CHANNEL_ID
        :param start_date: [OPTIONAL] (date|datetime) filter start date  [Defaults to UTC, if not given]
        :param end_date: [OPTIONAL] (date|datetime) filter end date  [Defaults to UTC, if not given]
        :param timezone: [OPTIONAL] timezone string from IAIA Time zone Database.  
            If provided it will be reflected in the given start/end values.
        :return: page
        """
        params = {}
        if start_date and end_date:
            if timezone:
                tz = pytz.timezone(timezone)
                if isinstance(start_date, datetime.date):
                    # make a datetime object
                    start_date = datetime.datetime.combine(start_date, datetime.datetime.min.time())
                if isinstance(end_date, datetime.date):
                    # make a datetime object
                    end_date = datetime.datetime.combine(end_date, datetime.datetime.min.time())
                start_date = start_date.replace(tzinfo=tz)
                end_date = end_date.replace(tzinfo=tz)

            params['start'] = start_date.strftime('%Y%m%d')  # YYYYMMDD
            params['end'] = end_date.strftime('%Y%m%d')  # YYYYMMDD
        if timezone:
            params['timezone'] = timezone

        url = 'https://{}/channels/{}/'.format(self._api_base_url,
                                               channel_id)
        response = self._session.get(url,
                                     params=params,
                                     auth=self._user_credentials)
        logger.debug('Response ({}): {}'.format(response.status_code, response.text))
        if response.status_code == 403:
            content = response.json()
            message = content.get('message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('StatusCode({}) not in {}: {}'.format(response.status_code,
                                                                     SUCCESSFUL_STATUS_CODES,
                                                                     response.text))
        page = response.json()
        yield page
        next_page_token = page.get('next_page_token', None)
        while 'next_page_token' in page and next_page_token:
            response = self._session.get(url,
                                         params={'next_page_token': next_page_token},
                                         auth=self._user_credentials)
            if response.status_code not in SUCCESSFUL_STATUS_CODES:
                raise UnknownError('StatusCode({}) not in {}: {}'.format(response.status_code,
                                                                         SUCCESSFUL_STATUS_CODES,
                                                                         response.text))
            page = response.json()
            yield page
            next_page_token = page['next_page_token']

    def search_channels(self, channel_ids=None, start_datetime_gte=None, end_datetime_lte=None, metadata_filters=None, orderby=None, reverse=False):
        """
        Provide Multi-channel Search & orderby features for the datalake API
        
        :param channel_ids: [OPTIONAL] (list) Channel_ids to filter results by
        :param start_datetime_gte: [OPTIONAL] (datetime object) 
        :param end_datetime_lte: [OPTIONAL] (datetime object) 
        :param metadata_filters: (dict) key/value of metadata fields to search by
            EXAMPLE:
                {
                    METADATA_FIELDNAME[CONDITIONAL]: FILTER_VALUE
                }
                WHERE:
                    METADATA_FIELDNAME:
                        'x-abeja-meta-FIELDNAME' value applied to the stored object.
                        NOTE: The 'x-abeja-meta-' prefix may be omitted.
                    CONDITIONAL:
                        __eq = EQUAL (==) [DEFAULT used if no conditional provided]
                        __gte = '>='
                        __lte = '<='
                        
        :param orderby: (list) metadata fields to order results by
        :param reverse: (bool) If True, and orderby fields given, the resulting order will be reversed 
            False [DEFAULT] == SMALL -> BIG [1, 2, 3, ... ]
            True == BIG -> Small [ ..., 3, 2, 1]
            
        :return: (dict) Filtered list of files
              {"files": [
                {
                  "file_id": "20170417T214948-3d0efe0c-6cf3-4919-9ca8-26bb07c861a9",
                  "download_url": "www.example.com",
                  "uploaded_at": "2017-04-17T21:49:48+00:00",
                  "content_type": "application/json",
                  "metadata": {
                    "x-abeja-meta-filename": "testfile"
                  }
                },
                ...
                ]}
        """
        if not start_datetime_gte:
            start_datetime_gte = ABEJA_UPLOADED_AT_EPOCH_UTC
        if not end_datetime_lte:
            end_datetime_lte = datetime.datetime.utcnow().replace(tzinfo=pytz.utc)

        # check that tzinfo is applied
        # --> If not, assume DEFAULT (UTC)
        if not start_datetime_gte.tzinfo:
            # default to UTC
            start_datetime_gte = pytz.utc.localize(start_datetime_gte)
            logger.warning('UTC applied to unaware datetime: {}'.format(start_datetime_gte))
        if not end_datetime_lte.tzinfo:
            # default to UTC
            end_datetime_lte = pytz.utc.localize(end_datetime_lte)
            logger.warning('UTC applied to unaware datetime: {}'.format(end_datetime_lte))

        # convert time to UTC (DEFAULT used in QUERY)
        start_datetime_gte = start_datetime_gte.astimezone(pytz.utc)
        end_datetime_lte = end_datetime_lte.astimezone(pytz.utc)

        logger.debug('start_datetime_gte: {}'.format(start_datetime_gte))
        logger.debug('end_datetime_lte: {}'.format(end_datetime_lte))

        # normalize metadata_filters
        valid_conditionals = ('eq', 'gte', 'lte')
        default_conditional = 'eq'
        normalized_metadata_filters = {}
        if metadata_filters:
            for k, v in metadata_filters.items():
                normalized_key = k.lower()
                if not k.endswith(valid_conditionals):
                    # set to default
                    normalized_key = '{}__{}'.format(normalized_key, default_conditional)
                if not k.startswith(ABEJA_METADATA_PREFIX):
                    # add 'x-abeja-meta-'
                    normalized_key = '{}{}'.format(ABEJA_METADATA_PREFIX, normalized_key)
                logger.debug('Normalizing metadata filter key({}) to: {}'.format(k, normalized_key))
                normalized_metadata_filters[normalized_key] = v
        logger.debug('Normalized metadata filters: {}'.format(normalized_metadata_filters))

        # generate date range for parallel requests
        request_date_ranges = []
        current = start_datetime_gte.date()
        while current <= end_datetime_lte.date():
            # NOTE: Currently creating single request per day... this can be tuned to reduce burden on our servers...
            start = current
            end = current
            request_date_ranges.append((start, end))
            current += datetime.timedelta(days=1)

        if not channel_ids:
            # collect channel ids
            channels = self.channels()
            channel_ids = [channel_info['channel_id'] for channel_info in channels]

        # prepare parallel request job parameters
        jobs = []
        for channel_id in channel_ids:
            url = 'https://{}/channels/{}/'.format(self._api_base_url,
                                                   channel_id)
            for start_date, end_date in request_date_ranges:
                job = (self._session,
                       self._user_credentials,
                       url,
                       start_date,
                       end_date)
                jobs.append(job)

        results = []
        pool = Pool(processes=REQUEST_PROCESSES)
        for request_args, request_url, response_pages in pool.imap_unordered(process_datalake_list_request, jobs):
            logger.debug('Request Processed: {}'.format(request_args))
            logger.debug('Request URL: {}'.format(request_url))
            for page in response_pages:
                for file_info in page['files']:
                    file_uploaded_at = file_info['uploaded_at']
                    # filter dates
                    if not (start_datetime_gte <= file_uploaded_at <= end_datetime_lte):
                        continue

                    file_metadata = file_info['metadata']
                    logger.debug('file metadata: {}'.format(file_metadata))
                    include_file = True
                    for filter_fieldname_condition, filter_value in normalized_metadata_filters.items():
                        # filter results
                        filter_metadata_fieldname, conditional = filter_fieldname_condition.rsplit('__', 1)
                        logger.debug(conditional)
                        try:
                            # raises a KeyError if key does not exist
                            file_metadata_value = file_metadata[filter_metadata_fieldname]
                            # Skip values if condition not met
                            if conditional == 'eq' and not (file_metadata_value == filter_value):
                                include_file = False
                                break
                            if conditional == 'gte' and not (file_metadata_value >= filter_value):
                                include_file = False
                                break
                            if conditional == 'lte' and not (file_metadata_value <= filter_value):
                                include_file = False
                                break
                        except KeyError:
                            logger.warning('File({}) missing metadata "{}", SKIPPING!'.format(file_info['file_id'], filter_metadata_fieldname))
                            include_file = False
                            break
                    if include_file:
                        results.append(file_info)

        if orderby:
            sortable_results = []
            unsortable_results = []
            # normalize metadata fieldnames
            normalized_orderby_fields = []
            for fieldname in orderby:
                normalized_fieldname = self._normalize_metadata_fieldname(fieldname)
                normalized_orderby_fields.append(normalized_fieldname)

            def keyby_fields(x):
                """
                Extract fields from x that are to be used for sorting
                :param x: (dict) file_info object
                :return: (list) field values to sort by
                """
                return [x['metadata'][field] for field in normalized_orderby_fields]

            for r in results:
                if all(f in r['metadata'] for f in normalized_orderby_fields):
                    sortable_results.append(r)
                else:
                    unsortable_results.append(r)

            sortable_results.sort(key=keyby_fields,
                                  reverse=reverse)
            results = sortable_results + unsortable_results

        return {'files': results}

