# -*- coding: utf-8 -*-
"""
abeja platform upload functions
"""
from __future__ import print_function, division, absolute_import, unicode_literals
import os
import logging
import datetime
import tarfile
import pytz
import requests
from collections import defaultdict
# for python2 support
try:
    from json.decoder import JSONDecodeError
    import tempfile  # use python 3 version
except ImportError:
    JSONDecodeError = ValueError
    from backports import tempfile

from time import sleep

from ..exceptions import Http403Forbidden, Http401Unauthorized, UnknownError, ResourceDoesNotExist, OrganizationNameNotSet
from ..core import PlatformResourceManager, retry, API_BASE_URL, SUCCESSFUL_STATUS_CODES
from .exceptions import InvalidModelHandler, InvalidModelBaseImage, InvalidModelProjectDirectory, ServiceStatusError, ServiceAliasIsNotUnique

logger = logging.getLogger('abejaapi')

MODELS_DATETIME_EPOCH_UTC = datetime.datetime(2017, 9, 11, tzinfo=pytz.utc)

VALID_CHANNEL_STORAGETYPES = (
    'datalake',
    'file',
    'rdb',
)

# http == MODEL API "service"
# event == MODEL API "trigger"
# backfill == MODEL API "run"
VALID_DEPLOYMENT_SERVICETYPES = (
    'http',
    'event',
    'backfill',
)

STORAGETYPE2SERVICENAME = {
    'rdb': 'datamart-rdb',
    'file': 'datamart-file',
}

VALID_SERVICE_STATUS = (
    'CREATE_COMPLETE',
    'UPDATE_COMPLETE',
    'CREATED'
)
ENTRIES_PER_PAGE = 50


class AbejaSettingsMissing(Exception):
    pass


class PlatformModelManager(PlatformResourceManager):
    """
    Provides features for interacting with data on the ABEJA Platform
    """

    def __init__(self, *args, **kwargs):
        # capture potentially set variables
        self._datalake_base_url = kwargs.pop('api_base_url', API_BASE_URL)
        self.project_directory = kwargs.pop('project_directory', None)

        # populate resource urls
        super(PlatformModelManager, self).__init__(*args, **kwargs)
        if not self._organization_name:
            # set to first available
            organization_name = [org_info['name'] for org_info in self.organizations()['organizations']][0]
            logger.debug('Setting Organization Name to: {}'.format(organization_name))
            self.set_organization(organization_name)

    def set_project_directory(self, directory):
        self.project_directory = directory

    @property
    def organization_endpoint(self):
        return 'https://{}/organizations/{}'.format(API_BASE_URL, self._organization_name)

    def models(self, model_id=None):
        """
        List Registered Models
        :return:
        """
        # TODO: Support paging
        url = '{}/models'.format(self.organization_endpoint)
        # NOTE: currently the API does not maintain the same interface with filter only
        # Response is different based on endpoint, using same endpoint for now with filtering, to get matching output.
        if model_id:
            url = '{}/{}'.format(url, model_id)

        response = self._session.get(url,
                                     auth=self._user_credentials)

        if response.status_code == 403:
            content = response.json()
            message = content.get('Message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code == 404 and 'not found' in response.text:
            raise ResourceDoesNotExist('[{}] StatusCode({}): {}'.format(url,
                                                                        response.status_code,
                                                                        response.text))
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('StatusCode({}) not in {}: {}'.format(response.status_code,
                                                                     SUCCESSFUL_STATUS_CODES,
                                                                     response.text))
        result = response.json()
        if model_id:
            result = {'entries': [result]}

        return result

    def detailed_models(self, model_id=None):
        """
        Provides a combined model information into a single call.
        Returned structure is a list of the following:
            MODEL:
                VERSIONS: [
                    SERVICES: []
                ]
        NOTE:
            'deployment_id' is added to displayed services
            'type' is added to displayed services: [ 'http', 'event', 'backfill']
                WHERE:
                    http     == MODEL API "service"
                    event    == MODEL API "trigger"
                    backfill == MODEL API "run"

            SERVICES includes services, runs, triggers.

        result = {
            'entries': [
                {
                    <model result>
                        'versions': [{
                            'deployment_id: xxx,
                            'deployment_name': yyy,
                            'services: [],
                        }],

                    },],
                }
            ]
        }
        :param model_id: (str) If given, only the given MODEL ID's related information will be returned.
        :return: (dict) {'entries':[{MODEL}, ... ] }
        """
        # get service_id to deployment_id mapping
        deployments = self.deployments()
        # to get all services we need to query all deployments
        all_services = []
        for deployment_info in deployments['entries']:
            deployment_id = deployment_info['deployment_id']
            deployment_services = self.services(deployment_id)
            endpoints = self.endpoints(deployment_id)
            for service_info in deployment_services['entries']:
                # add deployment_id
                service_info['deployment_id'] = deployment_id
                service_info['deployment_url'] = self.get_deployment_url(deployment_id)
                service_info['type'] = 'http'
                service_info['endpoints'] = []
                # find related endpoints
                for endpoint_info in endpoints['entries']:
                    if endpoint_info['service_alias'] == service_info['alias']:
                        service_info['endpoints'].append(endpoint_info)
                all_services.append(service_info)

            # add event triggers
            deployment_event_triggers = self.triggers(deployment_id)
            for event_trigger_info in deployment_event_triggers['entries']:
                event_trigger_info['deployment_id'] = deployment_id
                event_trigger_info['type'] = 'event'
                all_services.append(event_trigger_info)

            # add backfill runs
            deployment_backfill_runs = self.runs(deployment_id)
            for backfill_run_info in deployment_backfill_runs['entries']:
                backfill_run_info['deployment_id'] = deployment_id
                backfill_run_info['type'] = 'backfill'
                all_services.append(backfill_run_info)

        models = self.models(model_id)
        updated_entries = []
        for model_info in models['entries']:
            model_info['versions'] = []
            current_model_id = model_info['model_id']

            # get model versions
            versions = self.model_versions(current_model_id)
            for version_info in versions['entries']:
                version_info['services'] = []
                # Services linked by 'model_version_id'
                # -- Find services to include in versions
                for service_info in all_services:
                    if service_info['model_version_id'] == version_info['version_id']:
                        version_info['services'].append(service_info)
                model_info['versions'].append(version_info)
            updated_entries.append(model_info)
        result = {'entries': updated_entries}
        return result

    def add_model(self, name, description=''):
        """
        Adds a model to the Platform by creating a registration area.
        --> Creates
        :param name: (str) Model Name
        :param description: (str) Description of model purpose
        :return: (dict) JSON response to request
        """
        url = '{}/models'.format(self.organization_endpoint)
        payload = {
            'name': name,
            'description': description,
        }
        logger.debug('Request URL: {}'.format(url))
        logger.debug('Adding Model({})...'.format(name))
        response = self._session.post(url,
                                      json=payload,
                                      auth=self._user_credentials)
        logger.debug('response ({}): {}'.format(response.status_code,
                                                response.text))
        if response.status_code == 403:
            content = response.json()
            message = content.get('Message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('StatusCode({}) not in {}: {}'.format(response.status_code,
                                                                     SUCCESSFUL_STATUS_CODES,
                                                                     response.text))
        return response.json()

    def remove_cascade_model(self, model_id):
        """
        WARNIGN -- This method will remove the given model_id and all related dependencies.
        Delete order:
         1. remove_endpoints
         2. remove_services
         3. remove_deployments
         4. remove_versions
         5. remove_model

        :param model_id: (str) Model Identifier
        :return: (bool) True
        """
        # TODO: support paging of results
        results = {
            'action': 'remove-cascade-model',
            'responses': defaultdict(list)
         }
        deployments = self.deployments()
        for deployment_info in deployments['entries']:
            if deployment_info['model_id'] == model_id:
                deployment_id = deployment_info['deployment_id']

                endpoints = self.endpoints(deployment_id)
                for endpoint_info in endpoints['entries']:
                    endpoint_id = endpoint_info['endpoint_id']
                    logger.debug('DELETING endpoint({})...'.format(endpoint_id))
                    response = self.remove_endpoint(deployment_id, endpoint_id)
                    results['responses']['endpoints'].append(response)

                services = self.services(deployment_id)
                for service_info in services['entries']:
                    service_id = service_info['service_id']
                    logger.debug('DELETING service({})...'.format(service_id))
                    response = self.remove_service(deployment_id, service_id)
                    results['responses']['services'].append(response)

                triggers = self.triggers(deployment_id)
                for trigger_info in triggers['entries']:
                    trigger_id = trigger_info['trigger_id']
                    logger.debug('DELETING trigger({})...'.format(trigger_id))
                    response = self.remove_trigger(deployment_id)
                    results['responses']['triggers'].append(response)

                logger.debug('DELETING deployment({})...'.format(deployment_id))
                response = self.remove_deployment(deployment_id)
                results['responses']['deployments'].append(response)

        versions = self.model_versions(model_id)
        for version_info in versions['entries']:
            version_id = version_info['version_id']
            logger.debug('DELETING version({})...'.format(version_id))
            response = self.remove_model_version(model_id, version_id)
            results['responses']['versions'].append(response)

        logger.debug('DELETING model({})...'.format(model_id))
        response = self.remove_model(model_id)
        results['responses']['model'].append(response)

        return results

    def remove_model(self, model_id):
        """
        Remove (unregister?) the given model_id
        :param model_id: (str) Model Identifier
        :return: (dict) JSON response to request
        """
        url = '{}/models/{}'.format(self.organization_endpoint, model_id)
        response = self._session.delete(url,
                                        auth=self._user_credentials)
        if response.status_code == 403:
            content = response.json()
            message = content.get('Message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('[{}] StatusCode({}) not in {}: {}'.format(url,
                                                                          response.status_code,
                                                                          SUCCESSFUL_STATUS_CODES,
                                                                          response.text))
        return response.json()

    def _check_handler(self, user_defined_handler):
        """
        Confirms if the user_defined_handler is valid by checking that the function definition exists in the specified location.
        :param user_defined_handler: (str) Defines the function invoked when the deployed model is executed. In the form, "FILE:FUNCTION".
        :return: (booL) True (if valid nothing will happen and True returned)
        :raises: InvalidModelHandler
        """
        if ':' not in user_defined_handler:
            raise InvalidModelHandler('"{}" not in expected "FILE:FUNCTION" format!')

        if not self.project_directory or not os.path.exists(self.project_directory):
            raise InvalidModelProjectDirectory('Project Directory not set or does not exist: {}'.format(self.project_directory))

        module_fullname, function_name = user_defined_handler.split(':')
        module_parts = module_fullname.split('.')
        module_name = module_parts.pop(-1)
        module_filename = '{}.py'.format(module_name)
        module_basepath = os.path.join(self.project_directory, *module_parts)
        module_filepath = os.path.join(module_basepath, module_filename)

        # validate by checking for function name in module_filepath
        function_definition_startswith = 'def {}('.format(function_name)
        with open(module_filepath, 'r') as module_obj:
            content = module_obj.read()
            if function_definition_startswith not in content:
                raise InvalidModelHandler('Could not find function({}) in: {}'.format(function_name, module_filepath))
        return True

    def add_model_version(self, model_id, base_image, handler, version=None):
        """
        Register a model version and upload model to version
        :param model_id:
        :param deployment_id:
        :param base_image:
        :param version: (str)
        :param handler: (str) Defines the entry point (handler) to your function, in the form MODULE_NAME:FUNCTION_NAME
        :return:
        """
        if not version:
            version = '0.0.1'
        self._check_handler(handler)

        # 1. Register Version
        # -- Needed to retrieve target upload URL, obtained on version registration post request below
        url = '{}/models/{}/versions'.format(self.organization_endpoint, model_id)
        payload = {
            'image': base_image,
            'version': version,  # should be a version alias??!?
        }
        logger.debug('Request URL: {}'.format(url))

        response = self._session.post(url,
                                      json=payload,
                                      auth=self._user_credentials)
        logger.debug('response ({}): {}'.format(response.status_code,
                                                response.text))

        if response.status_code == 400 and 'not a valid image' in response.text:
            raise InvalidModelBaseImage("'{}' is not a valid image!".format(base_image))
        elif response.status_code == 403:
            content = response.json()
            message = content.get('Message', 'Reason Unknown')
            raise Http403Forbidden(message)

        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('StatusCode({}) not in {}: {}'.format(response.status_code,
                                                                     SUCCESSFUL_STATUS_CODES,
                                                                     response.text))
        version_info = response.json()

        # collect registered model assets into tar file for upload
        if not self.project_directory or not os.path.exists(self.project_directory):
            raise InvalidModelProjectDirectory('Project Directory not set or does not exist: {}'.format())
        project_name = os.path.basename(self.project_directory)
        tar_filename = '{}-{}.tar.gz'.format(project_name,
                                             version_info['version_id'])

        # directory and content gets cleaned up automatically with this context_manager
        with tempfile.TemporaryDirectory(prefix='abeja-models-') as temp_directory:
            tar_filepath = os.path.join(temp_directory, tar_filename)
            with tarfile.open(tar_filepath, 'w:gz') as archive:
                for item in os.listdir(self.project_directory):
                    item_path = os.path.join(self.project_directory, item)
                    if not item.startswith('.') and os.path.isfile(item_path):  # Exclude .git, etc
                        archive.add(item_path, arcname=item)

            # open file for sending
            with open(tar_filepath, 'rb') as model_assets_archive:
                response = requests.put(version_info['upload_url'],
                                        data=model_assets_archive)

        if response.status_code == 403:
            try:
                content = response.json()
                message = content.get('Message', 'Reason Unknown')
            except JSONDecodeError:
                message = response.text
            raise Http403Forbidden(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('StatusCode({}) not in {}: {}'.format(response.status_code,
                                                                     SUCCESSFUL_STATUS_CODES,
                                                                     response.text))
        return version_info

    def model_versions(self, model_id, version_id=None):
        url = '{}/models/{}/versions'.format(self.organization_endpoint, model_id)
        if version_id:
            url = '{}/{}'.format(url, version_id)
        response = self._session.get(url,
                                     auth=self._user_credentials)

        if response.status_code == 403:
            content = response.json()
            message = content.get('Message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('[{}] StatusCode({}) not in {}: {}'.format(url,
                                                                          response.status_code,
                                                                          SUCCESSFUL_STATUS_CODES,
                                                                          response.text))
        # versions return format is different from all others
        versions = response.json()
        # adjusting to match
        result = {'entries': versions}
        if version_id:
            result = {'entries': [versions]}
        return result

    def remove_model_version(self, model_id, version_id):  # TODO: Review, can model_id be removed here?!
        url = '{}/models/{}/versions/{}'.format(self.organization_endpoint, model_id, version_id)
        response = self._session.delete(url,
                                        auth=self._user_credentials)

        if response.status_code == 403:
            content = response.json()
            message = content.get('Message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('[{}] StatusCode({}) not in {}: {}'.format(url,
                                                                          response.status_code,
                                                                          SUCCESSFUL_STATUS_CODES,
                                                                          response.text))
        return response.json()  # TODO: Update model_api to provide better response, currently: {'message': 'ver-e2c28335717e4138 deleted'}

    def remove_cascade_model_version(self, model_id, version_id):
        """
        Remove version and it's related dependencies.
        NOTE: this will NOT delete the related deployment object, as it may contain multiple verions.
        :param model_id: (str) Model Identifier
        :param version_id: (str) Version Identifer
        :return: (dict) Dictionary containing a collection of the object IDs that were deleted.
        """
        results = {
            'action': 'remove-cascade-model-version',
            'responses': defaultdict(list)
         }
        # check that the version is correct, and exists
        versions = self.model_versions(model_id, version_id)
        assert len(versions['entries']) == 1
        version_info = versions['entries'][0]
        assert version_id == version_info['version_id']

        # get all deployments that use this version
        detailed_models = self.detailed_models(model_id)
        assert len(detailed_models['entries']) == 1
        detailed_model_info = detailed_models['entries'][0]
        for detailed_version_info in detailed_model_info['versions']:
            # get deployment_ids for this version from attached services
            if detailed_version_info['version_id'] == version_id:
                for detailed_service_info in detailed_version_info['services']:
                    deployment_id = detailed_service_info['deployment_id']

                    assert detailed_service_info['type'] in VALID_DEPLOYMENT_SERVICETYPES
                    if detailed_service_info['type'] == 'http':   # see VALID_DEPLOYMENT_SERVICETYPES
                        # delete related endpoints
                        for endpoint_info in detailed_service_info['endpoints']:
                            endpoint_id = endpoint_info['endpoint_id']
                            logger.debug('DELETING endpoint({})...'.format(endpoint_id))
                            response = self.remove_endpoint(deployment_id, endpoint_id)
                            results['responses']['endpoints'].append(response)

                        service_id = detailed_service_info['service_id']
                        logger.debug('DELETING service({})...'.format(service_id))
                        response = self.remove_service(deployment_id, service_id)
                        results['responses']['services'].append(response)

                    triggers = self.triggers(deployment_id)
                    for trigger_info in triggers['entries']:
                        trigger_id = trigger_info['trigger_id']
                        logger.debug('DELETING trigger({})...'.format(trigger_id))
                        response = self.remove_trigger(deployment_id)
                        results['responses']['triggers'].append(response)

        logger.debug('DELETING version({})...'.format(version_id))
        response = self.remove_model_version(model_id, version_id)
        results['responses']['versions'].append(response)
        return results

    def add_deployment(self, model_id, name=None):
        """
        Add a deployment entity to the given model
        :param name: (str) Name descriptor (If not given, name is generated in the format, 'model-MODELID_YYYYMMDDmmZ'
            WHERE:
                YYYYMMDDmmZ is the time now in UTC.
        :param model_id:
        :return: (dict) JSON Loaded Http Response
        """
        if not name:
            utc = datetime.datetime.utcnow()
            name = 'model-{}_{}Z'.format(model_id,
                                         utc.strftime('%Y%m%d%H%M'))
        url = '{}/models/{}/deployments'.format(self.organization_endpoint,
                                                model_id)
        payload = {
            'name': name,
        }
        response = self._session.post(url,
                                      json=payload,
                                      auth=self._user_credentials)
        if response.status_code == 403:
            content = response.json()
            message = content.get('Message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('[{}] StatusCode({}) not in {}: {}'.format(url,
                                                                          response.status_code,
                                                                          SUCCESSFUL_STATUS_CODES,
                                                                          response.text))
        return response.json()

    def deployments(self, deployment_id=None):
        url = '{}/deployments'.format(self.organization_endpoint)
        # NOTE: Workaround to align output atm...
        if deployment_id:
            url = '{}/{}'.format(url, deployment_id)

        has_next = True
        offset = 0
        params = {}  # for paging
        entries = []
        while has_next:
            response = self._session.get(url,
                                         params=params,
                                         auth=self._user_credentials)

            if response.status_code == 403:
                content = response.json()
                message = content.get('Message', 'Reason Unknown')
                raise Http403Forbidden(message)
            elif response.status_code not in SUCCESSFUL_STATUS_CODES:
                raise UnknownError('[{}] StatusCode({}) not in {}: {}'.format(url,
                                                                              response.status_code,
                                                                              SUCCESSFUL_STATUS_CODES,
                                                                              response.text))
            results = response.json()
            if deployment_id:
                has_next = False
            else:
                entries.extend(results['entries'])
                has_next = results.get('has_next', False)
                offset += 1
                params = {'offset': offset, 'limit': ENTRIES_PER_PAGE}

        if deployment_id:
            results = {'entries': [results]}
        return results

    def get_deployment_url(self, deployment_id):
        if not self._organization_name:
            raise OrganizationNameNotSet('Organization Name not set, use MANAGER.set_organization_name(NAME)!')
        url = "https://{}.{}/deployments/{}".format(self._organization_name, API_BASE_URL, deployment_id)
        return url

    def remove_deployment(self, deployment_id):
        url = '{}/deployments/{}'.format(self.organization_endpoint, deployment_id)
        response = self._session.delete(url,
                                        auth=self._user_credentials)

        if response.status_code == 403:
            content = response.json()
            message = content.get('Message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('[{}] StatusCode({}) not in {}: {}'.format(url,
                                                                          response.status_code,
                                                                          SUCCESSFUL_STATUS_CODES,
                                                                          response.text))
        return response.json()

    def add_service(self, deployment_id, version_id, alias=None):
        """

        :param deployment_id:
        :param version_id:
        :param alias: (str)  Matching '^[a-z0-9]+$'
        :return:
        """
        if not alias:
            # version_id = ver-NNNNNNNNN
            assert '-' in version_id
            _, version_integer = version_id.split('-')
            alias = version_integer
        alias = alias.lower().strip()

        url = '{}/deployments/{}/services'.format(self.organization_endpoint, deployment_id)

        payload = {
            'version_id': version_id,
            'alias': alias,
        }
        response = self._session.post(url,
                                      json=payload,
                                      auth=self._user_credentials)

        if response.status_code == 403:
            content = response.json()
            message = content.get('Message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code == 400 and 'Make sure alias is unique' in response.text:
            raise ServiceAliasIsNotUnique('Alias({}) is not unique!'.format(alias))
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('[{}] StatusCode({}) not in {}: {}'.format(url,
                                                                          response.status_code,
                                                                          SUCCESSFUL_STATUS_CODES,
                                                                          response.text))
        return response.json()

    def services(self, deployment_id, service_id=None):
        url = '{}/deployments/{}/services'.format(self.organization_endpoint, deployment_id)
        #if service_id:
        #    url = '{}/{}'.format(url, service_id)
        response = self._session.get(url,
                                     auth=self._user_credentials)

        if response.status_code == 403:
            content = response.json()
            message = content.get('Message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('[{}] StatusCode({}) not in {}: {}'.format(url,
                                                                          response.status_code,
                                                                          SUCCESSFUL_STATUS_CODES,
                                                                          response.text))
        services = response.json()
        result = {'entries': services}
        if service_id:
            for service_info in services:
                if service_info['service_id'] == service_id:
                    result = {'entries': [service_info]}
                    break
        return result

    def remove_service(self, deployment_id, service_id, remove_endpoints=False):
        if remove_endpoints:
            # remove related endpoints
            services = self.services(deployment_id, service_id)
            assert len(services['entries']) == 1
            service_info = services['entries'][0]
            service_alias = service_info['alias']
            # find related endpoints and delete
            endpoints = self.endpoints(deployment_id)
            for endpoint_info in endpoints['entries']:
                if endpoint_info['service_alias'] == service_alias:
                    endpoint_id = endpoint_info['endpoint_id']
                    self.remove_endpoint(deployment_id, endpoint_id)

        url = '{}/deployments/{}/services/{}'.format(self.organization_endpoint, deployment_id, service_id)
        response = self._session.delete(url,
                                        auth=self._user_credentials)

        if response.status_code == 403:
            content = response.json()
            message = content.get('Message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('[{}] StatusCode({}) not in {}: {}'.format(url,
                                                                          response.status_code,
                                                                          SUCCESSFUL_STATUS_CODES,
                                                                          response.text))
        return response.json()

    def _wait_for_valid_status(self, deployment_id, service_id, retries=100, sleep_seconds=12):
        """
        Check the given service status is VALID
        :param deployment_id: (str)
        :param service_id: (str)
        :param retries: (int)
        :param sleep_seconds: (int)
        :return:
        """

        # check that the service is in a valid staus
        service_status = None
        for retry in range(retries):
            services = self.services(deployment_id, service_id)
            service_info = services['entries'][0]
            service_status = service_info['status']
            if service_status in VALID_SERVICE_STATUS:
                break
            logger.debug('[{}] Service({}) with Invalid status({}), sleeping ({}) seconds...'.format(retry,
                                                                                                     service_id,
                                                                                                     service_status,
                                                                                                     sleep_seconds))
            sleep(sleep_seconds)

        if service_status not in VALID_SERVICE_STATUS:
            timeout_seconds = sleep_seconds * retries
            raise ServiceStatusError('Timeout({}s) ServiceStatus({}) not in functional state: {}'.format(timeout_seconds,
                                                                                                         service_status,
                                                                                                         VALID_SERVICE_STATUS))

    def add_endpoint(self, deployment_id, service_id, alias):  # TODO: Review, need deployment id?
        """
        An endpoint links an alias to a specific http service.
        This alias can be use
        NOTE: If the alias is set to 'default', the linked service can be accessed without setting the 'x-abeja-model-version-alias' http header.
        :param deployment_id:
        :param service_id:
        :param alias:
        :return:
        """
        # self._wait_for_valid_status(deployment_id, service_id)  # NOTE: issue with status ATM, commenting out for now
        # 500 - Internal Service Error occurs when add_endpoint is called soon after add_service,
        # In addition, there is an issue where status does not update from ACCEPTED.
        # As a workaround adding a simple wait here for now...
        sleep(8)

        url = "{}/deployments/{}/endpoints".format(self.organization_endpoint, deployment_id)

        payload = {
            'service_id': service_id,
            'custom_alias': alias,
        }
        response = self._session.post(url,
                                      json=payload,
                                      auth=self._user_credentials)

        if response.status_code == 403:
            content = response.json()
            message = content.get('Message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('[{}] StatusCode({}) not in {}: {}'.format(url,
                                                                          response.status_code,
                                                                          SUCCESSFUL_STATUS_CODES,
                                                                          response.text))
        return response.json()

    def endpoints(self, deployment_id, endpoint_id=None):
        url = '{}/deployments/{}/endpoints'.format(self.organization_endpoint, deployment_id)
        if endpoint_id:
            url = '{}/{}'.format(url, endpoint_id)
        response = self._session.get(url,
                                     auth=self._user_credentials)

        if response.status_code == 403:
            content = response.json()
            message = content.get('Message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code == 404 and '{} not found'.format(endpoint_id) in response.text:
            raise ResourceDoesNotExist('[{}] StatusCode({}): {}'.format(url,
                                                                        response.status_code,
                                                                        response.text))
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('[{}] StatusCode({}) not in {}: {}'.format(url,
                                                                          response.status_code,
                                                                          SUCCESSFUL_STATUS_CODES,
                                                                          response.text))
        endpoints = response.json()
        result = {'entries': endpoints}
        return result

    def update_endpoint(self, deployment_id, endpoint_id, service_id):
        """
        Update is used to change what service an endpoint points to
        :param deployment_id:
        :param endpoint_id:
        :param service_id:
        :return:
        """
        # self._wait_for_valid_status(deployment_id, service_id)  # NOTE: issue with status ATM, commenting out for now
        # 500 - Internal Service Error occurs when add_endpoint is called soon after add_service,
        # In addition, there is an issue where status does not update from ACCEPTED.
        # As a workaround adding a simple wait here for now...
        sleep(8)

        url = "{}/deployments/{}/endpoints/{}".format(self.organization_endpoint, deployment_id, endpoint_id)
        payload = {
            'service_id': service_id,
        }
        response = self._session.patch(url,
                                       json=payload,
                                       auth=self._user_credentials)

        if response.status_code == 403:
            content = response.json()
            message = content.get('Message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('[{}] StatusCode({}) not in {}: {}'.format(url,
                                                                          response.status_code,
                                                                          SUCCESSFUL_STATUS_CODES,
                                                                          response.text))
        return response.json()

    def remove_endpoint(self, deployment_id, endpoint_id):
        url = '{}/deployments/{}/endpoints/{}'.format(self.organization_endpoint, deployment_id, endpoint_id)
        response = self._session.delete(url,
                                        auth=self._user_credentials)

        if response.status_code == 403:
            content = response.json()
            message = content.get('Message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('[{}] StatusCode({}) not in {}: {}'.format(url,
                                                                          response.status_code,
                                                                          SUCCESSFUL_STATUS_CODES,
                                                                          response.text))
        return response.json()

    def add_run(self, deployment_id, version_id, input_operator, input_target, output_operator, output_target):
        """
        A 'run' is a single on-demand execution of a registered model deployment.

        :param deployment_id:
        :param version_id:
        :param input_operator:
        :param input_target:
        :param output_operator:
        :param output_target:
        :return:
        """
        url = "{}/deployments/{}/runs".format(self.organization_endpoint, deployment_id)

        # TODO: review! why are these input/output values different!?
        payload = {
            'version_id': version_id,
            'input_data': {input_operator: input_target},
            'output_template': {output_operator: output_target}
        }
        response = self._session.post(url,
                                      json=payload,
                                      auth=self._user_credentials)

        if response.status_code == 403:
            content = response.json()
            message = content.get('Message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('[{}] StatusCode({}) not in {}: {}'.format(url,
                                                                          response.status_code,
                                                                          SUCCESSFUL_STATUS_CODES,
                                                                          response.text))
        return response.json()

    def runs(self, deployment_id, run_id=None):
        url = '{}/deployments/{}/runs'.format(self.organization_endpoint, deployment_id)
        if run_id:
            url = '{}/{}'.format(url, run_id)
        response = self._session.get(url,
                                     auth=self._user_credentials)

        if response.status_code == 403:
            content = response.json()
            message = content.get('Message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('[{}] StatusCode({}) not in {}: {}'.format(url,
                                                                          response.status_code,
                                                                          SUCCESSFUL_STATUS_CODES,
                                                                          response.text))
        result = response.json()
        if run_id:
            result = {'entries': [result]}
        else:
            result = {'entries': result}
        return result

    def remove_run(self, deployment_id, run_id):
        raise NotImplementedError('In general this should not be necessary....')

    def _get_channel_storagetype(self, channel_id):
        storage_type = None
        result = self.channels()  # TODO: may need to include orgnization id
        for channel_info in result['channels']:
            if channel_id == channel_info['channel_id']:
                storage_type = channel_info['storage_type']
                break
        return storage_type

    def add_trigger(self, deployment_id, version_id, input_channel_id, output_channel_id):
        """
        Add an EVENT trigger.
        :param deployment_id:
        :param version_id:
        :param input_channel_id:
        :param output_channel_id:
        :return:
        """
        url = "{}/deployments/{}/triggers".format(self.organization_endpoint, deployment_id)

        input_storage_type = self._get_channel_storagetype(input_channel_id)
        output_storage_type = self._get_channel_storagetype(output_channel_id)
        input_service_name = STORAGETYPE2SERVICENAME[input_storage_type]
        output_service_name = STORAGETYPE2SERVICENAME[output_storage_type]

        # TODO: should name this (input|output)_channel_id
        # --> since we have the ID we shouldn't need service_name info anyway....
        payload = {
            'version_id': version_id,
            'input_service_name': input_service_name,
            'input_service_id': input_channel_id,
            'output_service_name': output_service_name,
            'output_service_id': output_channel_id,
        }
        response = self._session.post(url,
                                      json=payload,
                                      auth=self._user_credentials)

        if response.status_code == 403:
            content = response.json()
            message = content.get('Message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('[{}] StatusCode({}) not in {}: {}'.format(url,
                                                                          response.status_code,
                                                                          SUCCESSFUL_STATUS_CODES,
                                                                          response.text))
        return response.json()

    def triggers(self, deployment_id, trigger_id=None):
        url = '{}/deployments/{}/triggers'.format(self.organization_endpoint, deployment_id)
        if trigger_id:
            url = '{}/{}'.format(url, trigger_id)
        response = self._session.get(url,
                                     auth=self._user_credentials)

        if response.status_code == 403:
            content = response.json()
            message = content.get('Message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('[{}] StatusCode({}) not in {}: {}'.format(url,
                                                                          response.status_code,
                                                                          SUCCESSFUL_STATUS_CODES,
                                                                          response.text))
        result = response.json()
        if trigger_id:
            result = {'entries': [result]}
        else:
            result = {'entries': result}
        return result

    def remove_trigger(self, deployment_id, trigger_id):
        url = '{}/deployments/{}/triggers/{}'.format(self.organization_endpoint, deployment_id, trigger_id)
        response = self._session.delete(url,
                                        auth=self._user_credentials)

        if response.status_code == 403:
            content = response.json()
            message = content.get('Message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('[{}] StatusCode({}) not in {}: {}'.format(url,
                                                                          response.status_code,
                                                                          SUCCESSFUL_STATUS_CODES,
                                                                          response.text))
        return response.json()

    def service_logs(self, deployment_id, service_id, start_time=None, end_time=None):
        """
        View service logs
        :param deployment_id:
        :param service_id:
        :param start_time: (datetime object)
        :param end_time: (datetime object)
        :return:
        """
        # TODO: support next_token paging
        # when text_token=True, use offset: api?offset=1&limit=5
        # TODO: support no start/end
        url = "{}/deployments/{}/services/{}/logs".format(self.organization_endpoint, deployment_id, service_id)
        if not start_time:
            start_time = MODELS_DATETIME_EPOCH_UTC
        if not end_time:
            end_time = datetime.datetime.utcnow()
            end_time = end_time.replace(tzinfo=pytz.utc)
        params = {
            'start_time': start_time.isoformat(),
            'end_time': end_time.isoformat(),
        }
        response = self._session.get(url,
                                     params=params,
                                     auth=self._user_credentials)
        if response.status_code == 403:
            content = response.json()
            message = content.get('Message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('[{}] StatusCode({}) not in {}: {}'.format(url,
                                                                          response.status_code,
                                                                          SUCCESSFUL_STATUS_CODES,
                                                                          response.text))
        return response.json()

    def run_logs(self, deployment_id, run_id, start_time=None, end_time=None):
        """
        View run logs
        :param deployment_id:
        :param run_id:
        :param start_time: (datetime object)
        :param end_time: (datetime object)
        :return:
        """
        # TODO: support next_token paging
        # TODO: support no start/end
        url = "{}/deployments/{}/runs/{}/logs".format(self.organization_endpoint, deployment_id, run_id)
        if not start_time:
            start_time = MODELS_DATETIME_EPOCH_UTC
        if not end_time:
            end_time = datetime.datetime.utcnow()
            end_time = end_time.replace(tzinfo=pytz.utc)
        params = {
            'start_time': start_time.isoformat(),
            'end_time': end_time.isoformat(),
        }
        response = self._session.get(url,
                                     params=params,
                                     auth=self._user_credentials)
        if response.status_code == 403:
            content = response.json()
            message = content.get('Message', 'Reason Unknown')
            raise Http403Forbidden(message)
        elif response.status_code not in SUCCESSFUL_STATUS_CODES:
            raise UnknownError('[{}] StatusCode({}) not in {}: {}'.format(url,
                                                                          response.status_code,
                                                                          SUCCESSFUL_STATUS_CODES,
                                                                          response.text))
        return response.json()







